{
 "cells": [
  {
   "cell_type": "markdown",
   "metadata": {
    "collapsed": true,
    "deletable": true,
    "editable": true
   },
   "source": [
    "<img src=\"https://ibleducation.com/wp-content/uploads/2018/12/BrainLab-Rectangle-LightBG-Color@5x.png\" width=\"600\"/>"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {
    "deletable": true,
    "editable": true
   },
   "source": [
    "Welcome to BrainLab's hands-on training sessions! The purpose is to illustrate the concepts behind the demos in the course in a way that allows you to explore different functions, arguments and examples as you master the material."
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {
    "deletable": true,
    "editable": true
   },
   "source": [
    "## Recall: Blocks\n",
    "\n",
    "In the previous hands-on session, we learned that a block is made up of a block number, nonce, data field and hash. We understood that the hash encapsulated all of the other fields, and went through the mining process of finding a new nonce.\n",
    "\n",
    "We'll first recall all of what we learned about blocks and place the code in a neatly packaged Python `Block` object, which encapsulates our data and helper functions in one place. Later on, we'll define a `Blockchain` class that ties blocks together.\n",
    "\n",
    "Let's define the `Block` object, read it over, and dive into the differences with respect to the previous section."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 242,
   "metadata": {
    "collapsed": true,
    "deletable": true,
    "editable": true
   },
   "outputs": [],
   "source": [
    "import hashlib\n",
    "\n",
    "\n",
    "class Block:\n",
    "    \"\"\"Define a block class to package all relevant data and helper methods.\"\"\"\n",
    "    \n",
    "    def __init__(self, blockchain, is_first_block=False):\n",
    "        \"\"\"Initiate the block under the assumption that it must either be\n",
    "        (1) The first block on the chain, in which case its data will be prepopulated or\n",
    "        (2) Any other block, in which case we'll be able to derive some fields from the\n",
    "            last block on the chain.\n",
    "        \"\"\"\n",
    "        \n",
    "        if is_first_block:\n",
    "            # As the first block, its number will naturally be \"1\".\n",
    "            self.number = \"1\"\n",
    "            \n",
    "            # Its previous block's hash will simply consist of zeros (64 zeros, since\n",
    "            # that's the number of characters in a sha-256 hash), data field will be blank\n",
    "            # and a nonce of \"11316\" will be used to create a hash that meets the required\n",
    "            # level of difficulty.\n",
    "            self.prev_hash = \"0\" * 64\n",
    "            self.nonce = \"11316\"\n",
    "            self.set_data(\"\")\n",
    "\n",
    "        else:\n",
    "            # Since this is not the first block on the chain, we will need to derive some\n",
    "            # information from the last block on the current blockchain.\n",
    "            last_block = blockchain.get_last_block()\n",
    "            \n",
    "            # Increment the last block's number to get its own.\n",
    "            self.number = int(last_block.number) + 1\n",
    "            \n",
    "            # Base its last block's hash on the previous block's hash.\n",
    "            self.prev_hash = last_block.hash\n",
    "            \n",
    "            # Required fields, which will be populated via separate methods later.\n",
    "            self.nonce = None\n",
    "            self.data = None\n",
    "            self.hash = None\n",
    "        \n",
    "    \n",
    "    def set_data(self, data):\n",
    "        \"\"\"Update the data of the block as well as its own hash.\"\"\"\n",
    "        \n",
    "        self.data = data\n",
    "        self.set_hash()\n",
    "\n",
    "    \n",
    "    def set_prev_hash(self, prev_hash):\n",
    "        \"\"\"Update the previous hash field of the block as well as its own hash.\"\"\"\n",
    "        \n",
    "        self.prev_hash = prev_hash\n",
    "        self.set_hash()\n",
    "        \n",
    "\n",
    "    def set_hash(self):\n",
    "        \"\"\"Create a simple block's hash by concatenating all required fields.\"\"\"\n",
    "        \n",
    "        block_fields =  str(self.number) + \\\n",
    "                        str(self.nonce)  + \\\n",
    "                        str(self.data)   + \\\n",
    "                        str(self.prev_hash)\n",
    "        \n",
    "        # Use Python's native SHA-256 function.\n",
    "        self.hash = hashlib.sha256(block_fields.encode(\"utf-8\")).hexdigest()\n",
    "        \n",
    "        \n",
    "    def has_valid_hash(self):\n",
    "        \"\"\"Determine the validity of a hash, based on the number of zeros it starts with.\"\"\"\n",
    "        \n",
    "        difficulty = 4\n",
    "        return self.hash.startswith(\"0\" * difficulty)\n",
    "    \n",
    "        \n",
    "    def mine(self):\n",
    "        \"\"\"Obtain a nonce for the block that leads to a valid hash, considering its other fields.\"\"\"\n",
    "\n",
    "        # Start with a guess of \"1\", and increment it until we obtain a valid hash. At that point,\n",
    "        # store the last nonce value.\n",
    "        self.nonce = 1\n",
    "        \n",
    "        while self.nonce:\n",
    "            self.set_hash()\n",
    "\n",
    "            if self.has_valid_hash():\n",
    "                break\n",
    "\n",
    "            self.nonce += 1\n",
    "        \n",
    "    \n",
    "    def __str__(self):\n",
    "        \"\"\"Use to print the contents of the block in a helpful manner.\"\"\"\n",
    "        \n",
    "        output = \"Block |\"\n",
    "        \n",
    "        output += \"\\tNumber\\t\\t=>\\t\"   + str(self.number)     + \"\\n\"\n",
    "        output += \"\\tNonce\\t\\t=>\\t\"    + str(self.nonce)      + \"\\n\"\n",
    "        output += \"\\tData\\t\\t=>\\t\"     + str(self.data)       + \"\\n\"\n",
    "        output += \"\\tPrev. Hash\\t=>\\t\" + str(self.prev_hash)  + \"\\n\"\n",
    "        output += \"\\tHash\\t\\t=>\\t\"     + str(self.hash)\n",
    "        \n",
    "        return output"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {
    "deletable": true,
    "editable": true
   },
   "source": [
    "# Linking Blocks Together\n",
    "\n",
    "The main difference between this hands-on session and the previous one is that we now need to equip our blocks with the means to link to each other (after all, we've said that a \"blockchain\" is nothing more than a chain or list of blocks).\n",
    "\n",
    "## The Previous Hash Field\n",
    "\n",
    "If you look at the definition of a block above, you'll be able to spot the block fields that we've all come to know and love: the number, nonce, data and hash fields.\n",
    "\n",
    "But, there's an additional one at play here, which is called the \"previous hash field\" (`prev_hash` or `prev` for short). As you can see in the diagram below, it's what allows e.g. block #2 to know what the previous block in the chain is. Then, just like block #2 stores block #1's hash in its `prev` field, block #3 stores block #2's, and so on.\n",
    "\n",
    "<img src=\"https://ibleducation.com/wp-content/uploads/2018/12/brainlab-demo-blockchain.png\" width=\"600\" />\n",
    "\n",
    "Just like the block number, the previous hash field lets us order the blocks in the blockchain chronologically.\n",
    "\n",
    "### Diving into the Code\n",
    "\n",
    "The only real difference with respect to the last hands-on session is that our hashing function, `set_hash` under the `Block` object above, now takes into account the previous block's hash field in addition to the other fields we're familiar with.\n",
    "\n",
    "That's right! Our current block's hash *depends* on the previous block's hash, which means that it depends on the previous block's fields, which means that it depends on the previous block's previous block's hash, which means that it depends on those fields, and so on.\n",
    "\n",
    "Next, we'll see why that is, and why it's crucial for the integrity of the blockchain.\n",
    "\n",
    "<img src=\"https://ibleducation.com/wp-content/uploads/2018/12/brainlab-demo-blockchain-2.png\" width=\"600\" />\n",
    "\n",
    "## Why the Previous Hash Field Has to Be a Hash\n",
    "\n",
    "At first, it was not clear to me that the previous field needed to be the hash of the previous block. Why would this be necessary, if all we're trying to do is order blocks chronologically? Furthermore, doesn't the block number already provide the insight as to which block goes where on the chain? Is this `prev_hash` field redundant, then?\n",
    "\n",
    "\"No, young grasshopper\" — I tell my past self.\n",
    "\n",
    "Recall that the entire assumption behind the blockchain is that it must be **immutable**. We don't want to develop a chain of blocks under the assumption that nothing will ever change and then allow a third-party to just modify an old block... regardless of how small that modification is.\n",
    "\n",
    "That's where the concept of hashing comes again!\n",
    "\n",
    "By storing the hash of the previous block, we have the **certainty that nothing in the previous block will change**. If something did change in the previous block, then, as we know, its hash would change. At that point, we'd be able to compare it against the hash we had stored for it and detect that something fishy is going on.\n",
    "\n",
    "## A Note About the First Block\n",
    "\n",
    "You'll notice that, when we instantiate a new `Block`, the constructor asks if we're dealing with the first block in the blockchain.\n",
    "\n",
    "Needless to say, the first block in the chain will not have much to say about its previous block's hash field, since it won't exist.\n",
    "\n",
    "For that reason, we prepopulate its fields to have zeros for its previous hash, a block number of 1, an empty data field, and a nonce value `11316` in order to make sure that its hash still abides by our validity standards.\n",
    "\n",
    "# Assembling the Blockchain\n",
    "\n",
    "Now that we have a notion of the way in which blocks are linked on a chain via the `prev_hash` field, we're ready to assemble the code for a blockchain.\n",
    "\n",
    "We'll define a `Blockchain` object which, at its core, will contain a simple list and a series of helper functions to make our lives easier."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 214,
   "metadata": {
    "collapsed": true,
    "deletable": true,
    "editable": true
   },
   "outputs": [],
   "source": [
    "class Blockchain:\n",
    "    \"\"\"Define a blockchain class to package all relevant blocks and helper methods.\"\"\"\n",
    "    \n",
    "    def __init__(self):\n",
    "        \"\"\"Create an empty list to store our chain of blocks, or blockchain!\"\"\"\n",
    "        \n",
    "        self.blocks = []\n",
    "        \n",
    "        \n",
    "    def append_block(self, block):\n",
    "        \"\"\"Append a block object (defined above) to our chain of blocks.\"\"\"\n",
    "        \n",
    "        self.blocks.append(block)\n",
    "    \n",
    "        \n",
    "    def get_last_block(self):\n",
    "        \"\"\"Return the most recent block in the chain.\"\"\"\n",
    "        \n",
    "        return self.blocks[-1]\n",
    "    \n",
    "    \n",
    "    def validate(self):\n",
    "        \"\"\"Ensure that our blockchain is valid by:\n",
    "        (1) Verifying that every block has a valid hash. If an older block changes its\n",
    "            data and does not mine for a new nonce, then this check will fail.\n",
    "        (2) Verifying that every block's previous hash field links to the correct\n",
    "            previous block. If an older block is tampered with and its hash changes without\n",
    "            the knowledge of the next block in the chain, then this check will fail.\n",
    "        \"\"\"\n",
    "        \n",
    "        # In a chain of blocks A -> B -> C, this variable is used to compare C's previous\n",
    "        # hash field against B's own hash field.\n",
    "        expected_prev_hash = None\n",
    "        \n",
    "        # Iterate through the blocks, starting with the most recent one.\n",
    "        for block in reversed(self.blocks):\n",
    "            \n",
    "            # Check (1) above. Specify and output the number of the guilty block!\n",
    "            if not block.has_valid_hash():\n",
    "                print(\"Invalid blockchain due to invalid hash.\")\n",
    "                print(\"The guilty block is: #%s.\" % block.number)\n",
    "                break\n",
    "            \n",
    "            # Check (2) above. Specify and output the number of the guilty block!\n",
    "            if expected_prev_hash and (block.hash != expected_prev_hash):\n",
    "                print(\"Invalid blockchain due to mismatched prev_hash.\")\n",
    "                print(\"The guilty block is: #%s.\" % block.number)\n",
    "                break\n",
    "            \n",
    "            # Store the current block's previous hash as stated above.\n",
    "            expected_prev_hash = block.prev_hash\n",
    "        \n",
    "        # Conclude that our blockchain is valid if we didn't find any faulty blocks\n",
    "        # and, therefore, did not have to break out of the checking process.\n",
    "        else:       \n",
    "            print(\"Valid blockchain!\")\n",
    "    \n",
    "    \n",
    "    def __str__(self):\n",
    "        \"\"\"Use to print the contents of the blockchain in a helpful manner.\"\"\"\n",
    "        \n",
    "        output = \"\\n\\n\"\n",
    "        \n",
    "        for block in self.blocks:\n",
    "            output += block.__str__() + \"\\n\\n\\n\"\n",
    "        \n",
    "        return output"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {
    "deletable": true,
    "editable": true
   },
   "source": [
    "### Diving into the Code\n",
    "\n",
    "We've defined methods to perform the following tasks:\n",
    "\n",
    "- Initialize our list of blocks: `__init__(self)`\n",
    "- Append a new block object: `append_block(self, block)`\n",
    "- Get the last block: `get_last_block(self)`\n",
    "- Print the contents in a helpful manner: `__str__(self)`\n",
    "\n",
    "We've also define a validation method, called `validate(self)`, that performs an integrity check over the entire blockchain (starting with the last block) and alerts us if there's a block that doesn't meet our criteria.\n",
    "\n",
    "At this point, there are two checks that we're running on each block:\n",
    "\n",
    "1. To **verify that every block has a valid hash**. If an older block changes its data and does not mine for a new nonce, then this check will fail because its hash will not start with four zeros, for instance.\n",
    "\n",
    "2. To **verify that every block's previous hash field links to the correct previous block**. If an older block is tampered with and its hash changes without the knowledge of the next block on the chain, then this check will fail.\n",
    "\n",
    "At this point, we're ready to go through some examples.\n",
    "\n",
    "## Examples"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 243,
   "metadata": {
    "collapsed": false,
    "deletable": true,
    "editable": true
   },
   "outputs": [
    {
     "name": "stdout",
     "output_type": "stream",
     "text": [
      "\n",
      "\n",
      "Block |\tNumber\t\t=>\t1\n",
      "\tNonce\t\t=>\t11316\n",
      "\tData\t\t=>\t\n",
      "\tPrev. Hash\t=>\t0000000000000000000000000000000000000000000000000000000000000000\n",
      "\tHash\t\t=>\t000015783b764259d382017d91a36d206d0600e2cbb3567748f46a33fe9297cf\n",
      "\n",
      "\n",
      "\n"
     ]
    }
   ],
   "source": [
    "# Initialize our blockchain as an empty list of blocks.\n",
    "blockchain = Blockchain()\n",
    "\n",
    "# Create the first block with an empty data block and the required fields.\n",
    "# (inform the constructor that we're dealing with the first block).\n",
    "block1 = Block(blockchain, is_first_block=True)\n",
    "\n",
    "# Consequently, push the first onto the blockchain!\n",
    "blockchain.append_block(block1)\n",
    "\n",
    "# Let's print the blockchain to see our one block in action.\n",
    "print(blockchain)"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 244,
   "metadata": {
    "collapsed": false,
    "deletable": true,
    "editable": true
   },
   "outputs": [
    {
     "name": "stdout",
     "output_type": "stream",
     "text": [
      "\n",
      "\n",
      "Block |\tNumber\t\t=>\t1\n",
      "\tNonce\t\t=>\t11316\n",
      "\tData\t\t=>\t\n",
      "\tPrev. Hash\t=>\t0000000000000000000000000000000000000000000000000000000000000000\n",
      "\tHash\t\t=>\t000015783b764259d382017d91a36d206d0600e2cbb3567748f46a33fe9297cf\n",
      "\n",
      "\n",
      "Block |\tNumber\t\t=>\t2\n",
      "\tNonce\t\t=>\t298302\n",
      "\tData\t\t=>\tTx #318: 129899889\n",
      "\tPrev. Hash\t=>\t000015783b764259d382017d91a36d206d0600e2cbb3567748f46a33fe9297cf\n",
      "\tHash\t\t=>\t000005c7f16eb223212b02b7dbbc4526799ddec2bf39119447b467274bab0da6\n",
      "\n",
      "\n",
      "\n",
      "Valid blockchain!\n"
     ]
    }
   ],
   "source": [
    "# Now we're ready to add block #2 to our blockchain. We expect to see its\n",
    "# prev_hash field match the first block's.\n",
    "block2 = Block(blockchain)\n",
    "\n",
    "# We'll first set the data and then we'll mine to get a nonce that leads to\n",
    "# a valid hash. Lastly, we'll append it to the blockchain.\n",
    "block2.set_data('Tx #318: 129899889')\n",
    "block2.mine()\n",
    "blockchain.append_block(block2)\n",
    "\n",
    "# Let's see how our blockchain is looking, and validate it via our automated checks.\n",
    "print(blockchain)\n",
    "blockchain.validate()"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {
    "deletable": true,
    "editable": true
   },
   "source": [
    "Our blockchain is looking good so far, so let's add a third block and prepare to see what happens when we modify data that should not be modified in an immutable ledger!"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 245,
   "metadata": {
    "collapsed": false,
    "deletable": true,
    "editable": true
   },
   "outputs": [
    {
     "name": "stdout",
     "output_type": "stream",
     "text": [
      "\n",
      "\n",
      "Block |\tNumber\t\t=>\t1\n",
      "\tNonce\t\t=>\t11316\n",
      "\tData\t\t=>\t\n",
      "\tPrev. Hash\t=>\t0000000000000000000000000000000000000000000000000000000000000000\n",
      "\tHash\t\t=>\t000015783b764259d382017d91a36d206d0600e2cbb3567748f46a33fe9297cf\n",
      "\n",
      "\n",
      "Block |\tNumber\t\t=>\t2\n",
      "\tNonce\t\t=>\t298302\n",
      "\tData\t\t=>\tTx #318: 129899889\n",
      "\tPrev. Hash\t=>\t000015783b764259d382017d91a36d206d0600e2cbb3567748f46a33fe9297cf\n",
      "\tHash\t\t=>\t000005c7f16eb223212b02b7dbbc4526799ddec2bf39119447b467274bab0da6\n",
      "\n",
      "\n",
      "Block |\tNumber\t\t=>\t3\n",
      "\tNonce\t\t=>\t66286\n",
      "\tData\t\t=>\tOnce upon a time...\n",
      "\tPrev. Hash\t=>\t000005c7f16eb223212b02b7dbbc4526799ddec2bf39119447b467274bab0da6\n",
      "\tHash\t\t=>\t0000e7687a7b11f4512eca9aaa8e985049e42c946b00b5a47b0a4aaff15878f4\n",
      "\n",
      "\n",
      "\n"
     ]
    }
   ],
   "source": [
    "# Create and append block #3.\n",
    "block3 = Block(blockchain)\n",
    "block3.set_data('Once upon a time...')\n",
    "block3.mine()\n",
    "blockchain.append_block(block3)\n",
    "\n",
    "# How's our blockchain looking so far?\n",
    "print(blockchain)"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {
    "deletable": true,
    "editable": true
   },
   "source": [
    "Everything looks good, insofar as each block's previous hash field points to the last block, and our data is what we want it to be. What happens, though, if we decide to change the data field of Block #2?"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 246,
   "metadata": {
    "collapsed": false,
    "deletable": true,
    "editable": true
   },
   "outputs": [
    {
     "name": "stdout",
     "output_type": "stream",
     "text": [
      "\n",
      "\n",
      "Block |\tNumber\t\t=>\t1\n",
      "\tNonce\t\t=>\t11316\n",
      "\tData\t\t=>\t\n",
      "\tPrev. Hash\t=>\t0000000000000000000000000000000000000000000000000000000000000000\n",
      "\tHash\t\t=>\t000015783b764259d382017d91a36d206d0600e2cbb3567748f46a33fe9297cf\n",
      "\n",
      "\n",
      "Block |\tNumber\t\t=>\t2\n",
      "\tNonce\t\t=>\t298302\n",
      "\tData\t\t=>\tNew data for block #2!\n",
      "\tPrev. Hash\t=>\t000015783b764259d382017d91a36d206d0600e2cbb3567748f46a33fe9297cf\n",
      "\tHash\t\t=>\tc84bfcb8e34b4d65937fb2d62cb8b8fafa5cd9c9a2558530893342447b6ddc89\n",
      "\n",
      "\n",
      "Block |\tNumber\t\t=>\t3\n",
      "\tNonce\t\t=>\t66286\n",
      "\tData\t\t=>\tOnce upon a time...\n",
      "\tPrev. Hash\t=>\t000005c7f16eb223212b02b7dbbc4526799ddec2bf39119447b467274bab0da6\n",
      "\tHash\t\t=>\t0000e7687a7b11f4512eca9aaa8e985049e42c946b00b5a47b0a4aaff15878f4\n",
      "\n",
      "\n",
      "\n",
      "Invalid blockchain due to invalid hash.\n",
      "The guilty block is: #2.\n"
     ]
    }
   ],
   "source": [
    "block2.set_data('New data for block #2!')\n",
    "\n",
    "print(blockchain)\n",
    "blockchain.validate()"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {
    "deletable": true,
    "editable": true
   },
   "source": [
    "Aha! We changed the data field of block #2 and its hash changed to one that doesn't meet our validity standards (in other words, it doesn't start with four zeros).\n",
    "\n",
    "Well, no problem — we can re-mine block #2 and get a hash that does meet our standards! Let's see if that fixes our blockchain..."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 247,
   "metadata": {
    "collapsed": false,
    "deletable": true,
    "editable": true
   },
   "outputs": [
    {
     "name": "stdout",
     "output_type": "stream",
     "text": [
      "\n",
      "\n",
      "Block |\tNumber\t\t=>\t1\n",
      "\tNonce\t\t=>\t11316\n",
      "\tData\t\t=>\t\n",
      "\tPrev. Hash\t=>\t0000000000000000000000000000000000000000000000000000000000000000\n",
      "\tHash\t\t=>\t000015783b764259d382017d91a36d206d0600e2cbb3567748f46a33fe9297cf\n",
      "\n",
      "\n",
      "Block |\tNumber\t\t=>\t2\n",
      "\tNonce\t\t=>\t173283\n",
      "\tData\t\t=>\tNew data for block #2!\n",
      "\tPrev. Hash\t=>\t000015783b764259d382017d91a36d206d0600e2cbb3567748f46a33fe9297cf\n",
      "\tHash\t\t=>\t0000730421fb42485b90af10681499f8902579fc3568577e2fa9a553693885b4\n",
      "\n",
      "\n",
      "Block |\tNumber\t\t=>\t3\n",
      "\tNonce\t\t=>\t66286\n",
      "\tData\t\t=>\tOnce upon a time...\n",
      "\tPrev. Hash\t=>\t000005c7f16eb223212b02b7dbbc4526799ddec2bf39119447b467274bab0da6\n",
      "\tHash\t\t=>\t0000e7687a7b11f4512eca9aaa8e985049e42c946b00b5a47b0a4aaff15878f4\n",
      "\n",
      "\n",
      "\n",
      "Invalid blockchain due to mismatched prev_hash.\n",
      "The guilty block is: #2.\n"
     ]
    }
   ],
   "source": [
    "block2.mine()\n",
    "\n",
    "print(blockchain)\n",
    "blockchain.validate()"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {
    "deletable": true,
    "editable": true
   },
   "source": [
    "We were able to get a new hash for block #2, but our blockchain was still declared to be invalid by our automatic validator! Can you see why?\n",
    " \n",
    "It's because the hash that block #3 has in its `prev_hash` field does not match the hash that block #2 has for itself. Everybody's hash starts with four zeros as we'd expect, but that's not good enough because we know that a previous block has been modified.\n",
    "\n",
    "All we can do at this point is to reset block #3's `prev_hash` field and re-mine it (in other words, get a valid nonce) in order to make sure that our blockchain is valid again."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 248,
   "metadata": {
    "collapsed": false,
    "deletable": true,
    "editable": true
   },
   "outputs": [
    {
     "name": "stdout",
     "output_type": "stream",
     "text": [
      "\n",
      "\n",
      "Block |\tNumber\t\t=>\t1\n",
      "\tNonce\t\t=>\t11316\n",
      "\tData\t\t=>\t\n",
      "\tPrev. Hash\t=>\t0000000000000000000000000000000000000000000000000000000000000000\n",
      "\tHash\t\t=>\t000015783b764259d382017d91a36d206d0600e2cbb3567748f46a33fe9297cf\n",
      "\n",
      "\n",
      "Block |\tNumber\t\t=>\t2\n",
      "\tNonce\t\t=>\t173283\n",
      "\tData\t\t=>\tNew data for block #2!\n",
      "\tPrev. Hash\t=>\t000015783b764259d382017d91a36d206d0600e2cbb3567748f46a33fe9297cf\n",
      "\tHash\t\t=>\t0000730421fb42485b90af10681499f8902579fc3568577e2fa9a553693885b4\n",
      "\n",
      "\n",
      "Block |\tNumber\t\t=>\t3\n",
      "\tNonce\t\t=>\t20230\n",
      "\tData\t\t=>\tOnce upon a time...\n",
      "\tPrev. Hash\t=>\t0000730421fb42485b90af10681499f8902579fc3568577e2fa9a553693885b4\n",
      "\tHash\t\t=>\t0000f372ab112114f331a1dde72df62116bb72c7f12b04da2c5d20e72c347a9a\n",
      "\n",
      "\n",
      "\n",
      "Valid blockchain!\n"
     ]
    }
   ],
   "source": [
    "block3.set_prev_hash(block2.hash)\n",
    "block3.mine()\n",
    "\n",
    "print(blockchain)\n",
    "blockchain.validate()"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {
    "deletable": true,
    "editable": true
   },
   "source": [
    "At this point, we've restored our blockchain and learned that, in order to make it valid, every block's `prev_hash` field has to match its predecessor's `hash` field and have a valid hash of its own.\n",
    "\n",
    "We tampered with an old block in our blockchain (block #2), brought it to a halt, and learned that we could restore it simply by modifying and remining block #3.\n",
    "\n",
    "In the next hands-on session, we'll see how this wouldn't help us in the \"real world\", since modifying our own blockchain would have no impact on other people's distributed copies.\n",
    "\n",
    "# Next Steps: Distributed Blockchain!\n",
    "\n",
    "Now that we understand how blockchains works, we'll move onto why we need to have distributed copies of them."
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {
    "deletable": true,
    "editable": true
   },
   "source": [
    "<img src=\"https://ibleducation.com/wp-content/uploads/2018/12/BrainLab-Rectangle-LightBG-Color@5x.png\" width=\"600\"/>"
   ]
  }
 ],
 "metadata": {
  "kernelspec": {
   "display_name": "Python 3",
   "language": "python",
   "name": "python3"
  },
  "language_info": {
   "codemirror_mode": {
    "name": "ipython",
    "version": 3
   },
   "file_extension": ".py",
   "mimetype": "text/x-python",
   "name": "python",
   "nbconvert_exporter": "python",
   "pygments_lexer": "ipython3",
   "version": "3.5.1"
  }
 },
 "nbformat": 4,
 "nbformat_minor": 2
}
